""" flywheel """
try:
    from ._version import __version__  # pylint: disable=F0401
except ImportError:  # pragma: no cover
    __version__ = 'unknown'

import boto.dynamodb.types
from boto.dynamodb2.exceptions import ConditionalCheckFailedException
from boto.dynamodb.types import Binary
from boto.dynamodb2.types import (STRING, NUMBER, BINARY, STRING_SET,
                                  NUMBER_SET, BINARY_SET)
from decimal import Inexact, Rounded, Decimal


# HACK to force conversion of floats to Decimals
boto.dynamodb.types.DYNAMODB_CONTEXT.traps[Inexact] = False
boto.dynamodb.types.DYNAMODB_CONTEXT.traps[Rounded] = False


def float_to_decimal(f):  # pragma: no cover
    """ Monkey-patched replacement for boto's broken version """
    n, d = f.as_integer_ratio()
    numerator, denominator = Decimal(n), Decimal(d)
    ctx = boto.dynamodb.types.DYNAMODB_CONTEXT
    return ctx.divide(numerator, denominator)

boto.dynamodb.types.float_to_decimal = float_to_decimal


from .fields import Field, Composite, GlobalIndex, TypeDefinition, set_
from .models import Model
from .engine import Engine



""" Unit and system tests for flywheel """
import os

import inspect
import logging
import nose
import shutil
import subprocess
import tempfile
from boto.dynamodb2 import connect_to_region
from boto.dynamodb2.layer1 import DynamoDBConnection
from flywheel.engine import Engine
from urllib import urlretrieve


try:
    import unittest2 as unittest  # pylint: disable=F0401
except ImportError:
    import unittest


DYNAMO_LOCAL = 'https://s3-us-west-2.amazonaws.com/dynamodb-local/dynamodb_local_2013-12-12.tar.gz'


class DynamoLocalPlugin(nose.plugins.Plugin):

    """
    Nose plugin to run the Dynamo Local service

    See: http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Tools.html

    """
    name = 'dynamolocal'

    def __init__(self):
        super(DynamoLocalPlugin, self).__init__()
        self._dynamo_local = None
        self._dynamo = None
        self.port = None
        self.path = None
        self.link = None
        self.java = None
        self.region = None

    def options(self, parser, env):
        super(DynamoLocalPlugin, self).options(parser, env)
        tempdir = os.path.join(tempfile.gettempdir(), 'dynamolocal')
        parser.add_option('--dynamo-port', type=int, default=8000,
                          help="Run the Dynamo Local service on this port "
                          "(default 8000)")
        parser.add_option('--dynamo-path', default=tempdir,
                          help="Download the Dynamo Local server to this "
                          "directory (default %s)" % tempdir)
        parser.add_option('--dynamo-link', default=DYNAMO_LOCAL,
                          help="The link to the dynamodb local server code "
                          "(default %s)" % DYNAMO_LOCAL)
        parser.add_option('--dynamo-java', default='java',
                          help="The path to the java executable (default java)")
        parser.add_option('--dynamo-region', help="If provided, tests will "
                          "create tables in this LIVE dynamo region "
                          "instead of using DynamoDB Local. USE WITH CAUTION.")

    def configure(self, options, conf):
        super(DynamoLocalPlugin, self).configure(options, conf)
        self.port = options.dynamo_port
        self.path = options.dynamo_path
        self.link = options.dynamo_link
        self.java = options.dynamo_java
        self.region = options.dynamo_region
        logging.getLogger('boto').setLevel(logging.WARNING)

    @property
    def dynamo(self):
        """ Lazy loading of the dynamo connection """
        if self._dynamo is None:
            if self.region is not None:
                self._dynamo = connect_to_region(self.region)
            else:
                if not os.path.exists(self.path):
                    tarball = urlretrieve(self.link)[0]
                    subprocess.check_call(['tar', '-zxf', tarball])
                    name = os.path.basename(self.link).split('.')[0]
                    shutil.move(name, self.path)
                    os.unlink(tarball)

                lib_path = os.path.join(self.path, 'DynamoDBLocal_lib')
                jar_path = os.path.join(self.path, 'DynamoDBLocal.jar')
                cmd = [self.java, '-Djava.library.path=' + lib_path, '-jar',
                       jar_path, '--port', str(self.port)]
                self._dynamo_local = subprocess.Popen(cmd,
                                                      stdout=subprocess.PIPE,
                                                      stderr=subprocess.STDOUT)
                self._dynamo = DynamoDBConnection(
                    host='localhost',
                    port=self.port,
                    is_secure=False,
                    aws_access_key_id='',
                    aws_secret_access_key='')
        return self._dynamo

    def startContext(self, context):  # pylint: disable=C0103
        """ Called at the beginning of modules and TestCases """
        # If this is a TestCase, dynamically set the dynamo connection
        if inspect.isclass(context) and hasattr(context, 'dynamo'):
            context.dynamo = self.dynamo

    def finalize(self, result):
        """ terminate the dynamo local service """
        if self._dynamo_local is not None:
            self._dynamo_local.terminate()
            if not result.wasSuccessful():
                print self._dynamo_local.stdout.read()


class DynamoSystemTest(unittest.TestCase):

    """ Base class for tests that need an :class:`~flywheel.engine.Engine` """
    dynamo = None
    models = []

    @classmethod
    def setUpClass(cls):
        super(DynamoSystemTest, cls).setUpClass()
        logging.getLogger('boto').setLevel(logging.WARNING)
        cls.engine = Engine(cls.dynamo, ['test'])
        cls.engine.register(*cls.models)
        cls.engine.create_schema()

    @classmethod
    def tearDownClass(cls):
        super(DynamoSystemTest, cls).tearDownClass()
        cls.engine.delete_schema()

    def tearDown(self):
        super(DynamoSystemTest, self).tearDown()
        for model in self.engine.models.itervalues():
            self.engine.scan(model).delete()



""" Model metadata and metaclass objects """
import time

import inspect
from boto.dynamodb2.fields import HashKey, RangeKey, BaseSchemaField
from boto.dynamodb2.table import Table
from boto.exception import JSONResponseError
from collections import defaultdict

from .fields import Field


class ValidationError(Exception):

    """ Model inconsistency """
    pass


class Ordering(object):

    """
    A way that the models are ordered

    This will be a combination of a hash key and a range key. It may be the
    primary key, a local secondary index, or a global secondary index.

    """

    def __init__(self, meta, hash_key, range_key=None, index_name=None):
        self.meta = meta
        self.hash_key = hash_key
        self.range_key = range_key
        self.index_name = index_name

    def query_kwargs(self, **kwargs):
        """ Get the boto query kwargs for querying against this index """
        kwargs = {'%s__eq' % self.hash_key.name:
                  self.hash_key.resolve(scope=kwargs)}
        if self.index_name is not None:
            kwargs['index'] = self.index_name
        return kwargs

    def __repr__(self):
        if self.range_key is None:
            return "Ordering(%s, None, %s)" % (self.hash_key.name,
                                               self.index_name)
        else:
            return "Ordering(%s, %s, %s)" % (self.hash_key.name,
                                             self.range_key.name,
                                             self.index_name)


def merge_metadata(cls):
    """
    Merge all the __metadata__ dicts in a class's hierarchy

    keys that do not begin with '_' will be inherited.

    keys that begin with '_' will only apply to the object that defines them.

    """
    cls_meta = cls.__dict__.get('__metadata__', {})
    meta = {}
    for base in cls.__bases__:
        meta.update(getattr(base, '__metadata__', {}))
    # Don't merge any keys that start with '_'
    for key in meta.keys():
        if key.startswith('_'):
            del meta[key]
    meta.update(cls_meta)
    return meta


class ModelMetaclass(type):

    """
    Metaclass for Model objects

    Merges model metadata, sets the ``meta_`` attribute, and performs
    validation checks.

    """
    def __new__(mcs, name, bases, dct):
        cls = super(ModelMetaclass, mcs).__new__(mcs, name, bases, dct)

        cls.__metadata__ = merge_metadata(cls)
        cls.__on_create__()

        cls.meta_ = cls.__metadata_class__(cls)
        cls.meta_.post_create()
        cls.meta_.validate_model()
        cls.meta_.post_validate()

        cls.__after_create__()

        return cls

    def __init__(cls, name, bases, dct):
        super(ModelMetaclass, cls).__init__(name, bases, dct)


class ModelMetadata(object):

    """
    Container for model metadata

    Parameters
    ----------
    model : :class:`.Model`

    Attributes
    ----------
    name : str
        The unique name of the model. This is set by the '_name' field in
        __metadata__. Defaults to the name of the model class.
    abstract : bool
        If a model is abstract then it has no table in Dynamo
    namespace : list
        The namespace of this model. Set by the Engine.
    global_indexes : list
        List of global indexes (hash_key, [range_key]) pairs.
    related_fields : dict
        Mapping of field names to set of fields that change when that field
        changes (usually just that field name, but can be more if composite
        fields use it)
    orderings : list
        List of :class:`.Ordering`
    throughput : dict
        Mapping of 'read' and 'write' to the table throughput (default 5, 5)

    """
    __order_class__ = Ordering
    namespace = []

    def __init__(self, model):
        self.model = model
        self._name = model.__name__
        self.global_indexes = []
        self.orderings = []
        self.throughput = {'read': 5, 'write': 5}
        self._abstract = False
        self.__dict__.update(model.__metadata__)
        self.name = self._name
        self.fields = {}
        self.hash_key = None
        self.range_key = None
        self.related_fields = defaultdict(set)
        self.all_global_indexes = set()
        for gindex in self.global_indexes:
            self.all_global_indexes.add(gindex.hash_key)
            if gindex.range_key is not None:
                self.all_global_indexes.add(gindex.range_key)

    def post_create(self):
        """ Create the orderings """
        for name, member in inspect.getmembers(self.model):
            if isinstance(member, Field):
                self.fields[name] = member
                member.name = name
                member.model = self.model
                if member.hash_key:
                    self.hash_key = member
                elif member.range_key:
                    self.range_key = member

        self.orderings.append(self.__order_class__(self, self.hash_key,
                                                   self.range_key))
        for field in self.fields.itervalues():
            if field.index:
                order = self.__order_class__(self, self.hash_key, field,
                                             field.index_name)
                self.orderings.append(order)

        for index in self.global_indexes:
            for key in index:
                if key not in self.fields:
                    raise ValidationError("Model %s global index %s "
                                          "references unknown field '%s'" %
                                          (self.name, index.name, key))
            if index.range_key:
                range_key = self.fields[index.range_key]
            else:
                range_key = None
            self.orderings.append(self.__order_class__(self,
                                                       self.fields[
                                                           index.hash_key],
                                                       range_key, index.name))

    def post_validate(self):
        """ Build the dict of related fields """
        def update_related(field, name):
            """ Recursively add a field to related """
            for f in field.subfields:
                self.related_fields[f].add(name)
                subfield = self.fields[f]
                if subfield.composite:
                    update_related(subfield, name)

        for field in self.fields.itervalues():
            self.related_fields[field.name].add(field.name)
            if field.composite:
                update_related(field, field.name)

    def get_ordering_from_fields(self, eq_fields, fields):
        """
        Get a unique ordering from constraint fields

        Parameters
        ----------
        eq_fields : list
        fields : list

        Returns
        -------
        ordering : :class:`.Ordering`

        Raises
        ------
        exc : :class:`TypeError`
            If more than one possible Ordering is found

        """
        orderings = []
        for order in self.orderings:
            needed = order.hash_key.can_resolve(eq_fields)
            if len(needed) == 0:
                continue

            # If there are no non-equality fields, range key must be in the
            # eq_fields
            if len(fields) == 0:
                rng_fields = set(eq_fields)
                rng_needed = order.range_key.can_resolve(rng_fields)
                # hash and range key must use all eq_fields
                if len(set(eq_fields) - needed - rng_needed) != 0:
                    continue
            else:
                # If there are eq_fields left over, continue
                if len(set(eq_fields) - needed) != 0 or len(fields) > 1:
                    continue
                if order.range_key.name not in fields:
                    continue

            orderings.append(order)

        if len(orderings) > 1:
            for order in orderings:
                if order.index_name is None:
                    return order
            raise ValueError("More than one ordering found: %s" % orderings)
        elif len(orderings) == 1:
            return orderings[0]
        else:
            return None

    def get_ordering_from_index(self, index):
        """ Get the ordering with matching index name """
        for order in self.orderings:
            if order.index_name == index:
                return order
        raise ValueError("Cannot find ordering with index name '%s'" % index)

    def rk(self, obj=None, scope=None):
        """ Construct the range key value """
        if self.range_key is not None:
            return self.range_key.resolve(obj, scope)

    def hk(self, obj=None, scope=None):
        """ Construct the primary key value """
        return self.hash_key.resolve(obj, scope)

    def pk_dict(self, obj=None, scope=None, ddb_dump=False):
        """ Get the dynamo primary key dict for an item """
        # If we can unambiguously tell that a single string defines the primary
        # key, allow scope to be a single string
        if (obj is None and isinstance(scope, basestring) and
                self.range_key is None):
            scope = {self.hash_key.name: scope}

        hk = self.hk(obj, scope)
        if ddb_dump:
            hk = self.hash_key.ddb_dump(hk)
        rk = self.rk(obj, scope)
        key_dict = {self.hash_key.name: hk}
        if rk is not None:
            if ddb_dump:
                rk = self.range_key.ddb_dump(rk)
            key_dict[self.range_key.name] = rk
        return key_dict

    @property
    def abstract(self):
        """ Getter for abstract """
        return self._abstract

    @property
    def ddb_tablename(self):
        """ The name of the DynamoDB table """
        if self.abstract:
            return None
        return '-'.join(self.namespace + [self.name])

    def ddb_table(self, connection):
        """ Construct a Dynamo table from a connection """
        if self.abstract:
            return None
        return Table(self.ddb_tablename, connection=connection)

    def validate_model(self):
        """ Perform validation checks on the model declaration """
        if self.abstract or self.model.__dict__.get('__abstract__'):
            return
        hash_keys = [f for f in self.fields.values() if f.hash_key]
        range_keys = [f for f in self.fields.values() if f.range_key]
        indexes = [f for f in self.fields.values() if f.index]
        name = self.name
        if len(hash_keys) != 1:
            raise ValidationError("Model %s must have exactly one hash key" %
                                  name)
        if len(range_keys) > 1:
            raise ValidationError("Model %s can't have more than one range key"
                                  % name)
        if len(indexes) > 5:
            raise ValidationError("Model %s can't have more than 5 "
                                  "local indexes" % name)
        if len(indexes) > 0 and len(range_keys) == 0:
            raise ValidationError("Model %s can't set indexes without a "
                                  "range key" % name)
        if len(self.global_indexes) > 5:
            raise ValidationError("Model %s can't have more than 5 "
                                  "global indexes" % name)
        for field in self.fields.values():
            if field.composite:
                for f in field.subfields:
                    if f not in self.fields:
                        raise ValidationError("Model %s key %s references "
                                              "unknown field '%s'" %
                                              (name, field.name, f))
                    if f == field.name:
                        raise ValidationError("Model %s key %s cannot contain "
                                              "itself" % (name, field.name))

    def create_dynamo_schema(self, connection, tablenames=None, test=False,
                             wait=False, throughput=None):
        """
        Create all Dynamo tables for this model

        Parameters
        ----------
        connection : :class:`~boto.dynamodb2.layer1.DynamoDBConnection`
        tablenames : list, optional
            List of tables that already exist. Will call 'describe' if not
            provided.
        test : bool, optional
            If True, don't actually create the table (default False)
        wait : bool, optional
            If True, block until table has been created (default False)
        throughput : dict, optional
            The throughput of the table and global indexes. Has the keys 'read'
            and 'write'. To specify throughput for global indexes, add the name
            of the index as a key and another 'read', 'write' dict as the
            value.

        Returns
        -------
        table : str
            Table name that was created, or None if nothing created

        """
        if self.abstract:
            return None
        if tablenames is None:
            tablenames = connection.list_tables()['TableNames']
        if self.ddb_tablename in tablenames:
            return None
        elif test:
            return self.ddb_tablename

        attrs = []
        indexes = []
        global_indexes = []
        hash_key = None
        raw_attrs = {}

        if throughput is not None:
            table_throughput = throughput
        else:
            table_throughput = self.throughput
        table_throughput = {
            'ReadCapacityUnits': table_throughput['read'],
            'WriteCapacityUnits': table_throughput['write'],
        }

        hash_key = HashKey(self.hash_key.name,
                           data_type=self.hash_key.ddb_data_type)
        schema = [hash_key.schema()]
        for name, field in self.fields.iteritems():
            if field.hash_key:
                f = hash_key
            elif field.range_key:
                f = RangeKey(name, data_type=field.ddb_data_type)
                schema.append(f.schema())
            elif field.index:
                idx = field.get_boto_index(hash_key)
                f = idx.parts[1]
                indexes.append(idx.schema())
            elif any(map(lambda x: name in x, self.global_indexes)):
                f = BaseSchemaField(name, data_type=field.ddb_data_type)
            else:
                continue
            attrs.append(f.definition())
            raw_attrs[name] = f

        for gindex in self.global_indexes:
            index = gindex.get_boto_index(self.fields)
            if throughput is not None and gindex.name in throughput:
                index.throughput = throughput[gindex.name]
            global_indexes.append(index.schema())

        # Make sure indexes & global indexes either have data or are None
        indexes = indexes or None
        global_indexes = global_indexes or None
        if not test:
            connection.create_table(
                attrs, self.ddb_tablename, schema, table_throughput,
                local_secondary_indexes=indexes,
                global_secondary_indexes=global_indexes)
            if wait:
                desc = connection.describe_table(self.ddb_tablename)
                while desc['Table']['TableStatus'] != 'ACTIVE':
                    time.sleep(1)
                    desc = connection.describe_table(self.ddb_tablename)

        return self.ddb_tablename

    def delete_dynamo_schema(self, connection, tablenames=None, test=False,
                             wait=False):
        """
        Drop all Dynamo tables for this model

        Parameters
        ----------
        connection : :class:`~boto.dynamodb2.layer1.DynamoDBConnection`
        tablenames : list, optional
            List of tables that already exist. Will call 'describe' if not
            provided.
        test : bool, optional
            If True, don't actually delete the table (default False)
        wait : bool, optional
            If True, block until table has been deleted (default False)

        Returns
        -------
        table : str
            Table name that was deleted, or None if nothing deleted

        """
        if self.abstract:
            return None
        if tablenames is None:
            tablenames = connection.list_tables()['TableNames']

        if self.ddb_tablename in tablenames:
            if not test:
                self.ddb_table(connection).delete()
                if wait:
                    try:
                        connection.describe_table(self.ddb_tablename)
                        while True:
                            time.sleep(1)
                            connection.describe_table(self.ddb_tablename)
                    except JSONResponseError as e:
                        if e.status != 400:
                            raise
            return self.ddb_tablename
        return None



""" Field type definitions """
import functools
import datetime

import json
import boto.s3.key
from boto.dynamodb.types import Binary, float_to_decimal
from boto.dynamodb2.types import (NUMBER, STRING, BINARY, NUMBER_SET,
                                  STRING_SET, BINARY_SET)
from decimal import Decimal


ALL_TYPES = {}


def set_(data_type):
    """ Create an alias for a SetType that contains this data type """
    return frozenset([data_type])


def register_type(type_class, allow_in_set=True):
    """ Register a type class for use with Fields """
    ALL_TYPES[type_class.data_type] = type_class
    if allow_in_set:
        set_key = set_(type_class.data_type)
        ALL_TYPES[set_key] = SetType.bind(type_class.data_type)
    for alias in type_class.aliases:
        ALL_TYPES[alias] = type_class
        ALL_TYPES[set_(alias)] = SetType.bind(alias)


class TypeDefinition(object):

    """
    Base class for all Field types

    Attributes
    ----------
    data_type : object
        The value you wish to pass in to Field as the data_type.
    aliases : list
        Other values that will reference this type if passed to Field
    ddb_data_type : {STRING, BINARY, NUMBER, STRING_SET, BINARY_SET, NUMBER_SET}
        The DynamoDB data type that backs this type
    mutable : bool
        If True, flywheel will track updates to this field automatically when
        making calls to sync()
    allowed_filters : set
        The set of filters that can be used on this field type

    """
    data_type = None
    aliases = []
    ddb_data_type = None
    mutable = False

    def __init__(self):
        if self.ddb_data_type == NUMBER:
            self.allowed_filters = set(['eq', 'ne', 'lte', 'lt', 'gte', 'gt',
                                        'in', 'between'])
        elif self.ddb_data_type in (STRING, BINARY):
            self.allowed_filters = set(['eq', 'ne', 'lte', 'lt', 'gte', 'gt',
                                        'in', 'between', 'beginswith'])
        elif self.ddb_data_type in (NUMBER_SET, STRING_SET, BINARY_SET):
            self.allowed_filters = set(['contains', 'ncontains', 'in'])
        else:
            raise ValueError("Unknown dynamo data type '%s'" %
                             self.ddb_data_type)

    def coerce(self, value, force):
        """
        Check the type of a value and possible convert it

        Parameters
        ----------
        value : object
            The value to check
        force : bool
            If True, always attempt to convert a bad type to the correct type

        Returns
        -------
        value : object
            A variable of the correct type

        Raises
        ------
        exc : TypeError or ValueError
            If the value is the incorrect type and could not be converted

        """

        if not isinstance(value, self.data_type):
            raise TypeError()
        return value

    def ddb_dump_inner(self, value):
        """
        If this is a set type, dump a value to the type contained in the set

        """
        return value

    def ddb_dump(self, value):
        """ Dump a value to a form that can be stored in DynamoDB """
        return value

    def ddb_load(self, value):
        """ Turn a value into this type from a DynamoDB value """
        return value

    def __repr__(self):
        return u'TypeDefinition(%s)' % self

    def __unicode__(self):
        return unicode(self.data_type)

    def __str__(self):
        return unicode(self).encode('utf-8')


class SetType(TypeDefinition):

    """ Set types """
    data_type = set
    mutable = True

    def __init__(self, item_type=None, type_class=None):
        self.item_type = item_type
        set_map = {
            STRING: STRING_SET,
            NUMBER: NUMBER_SET,
            BINARY: BINARY_SET,
        }
        if item_type is not None and type_class is None:
            self.item_field = ALL_TYPES[item_type]()
            self.ddb_data_type = set_map[self.item_field.ddb_data_type]
        elif type_class is not None:
            self.item_field = type_class()
            self.ddb_data_type = set_map[self.item_field.ddb_data_type]
        else:
            self.item_field = None
            self.ddb_data_type = STRING_SET
        super(SetType, self).__init__()

    def coerce(self, value, force):
        if not isinstance(value, set):
            if force:
                value = set(value)
            else:
                raise TypeError()
        if self.item_field is not None:
            converted_values = set()
            for item in value:
                converted_values.add(self.item_field.coerce(item, force))
            return converted_values
        return value

    def ddb_dump_inner(self, value):
        """ We need to expose this for 'contains' and 'ncontains' """
        if self.item_field is None:
            return value
        return self.item_field.ddb_dump(value)

    def ddb_dump(self, value):
        if self.item_field is None:
            return value
        return set([self.ddb_dump_inner(v) for v in value])

    def ddb_load(self, value):
        if self.item_field is None:
            return value
        return set([self.item_field.ddb_load(v) for v in value])

    def __unicode__(self):
        if self.item_type is None:
            return super(SetType, self).__unicode__()
        else:
            return unicode(set([self.item_type]))

    @classmethod
    def bind(cls, item_type):
        """ Create a set factory that will contain a specific data type """
        return functools.partial(cls, item_type)

register_type(SetType, allow_in_set=False)
ALL_TYPES[STRING_SET] = SetType.bind(STRING)
ALL_TYPES[BINARY_SET] = SetType.bind(BINARY)
ALL_TYPES[NUMBER_SET] = SetType.bind(NUMBER)


class NumberType(TypeDefinition):

    """ Any kind of numerical value """
    data_type = NUMBER
    ddb_data_type = NUMBER

    def coerce(self, value, force):
        if not (isinstance(value, float) or isinstance(value, int) or
                isinstance(value, long) or isinstance(value, Decimal)):
            if force:
                try:
                    return int(value)
                except ValueError:
                    return float(value)
            else:
                raise TypeError()
        return value

    def ddb_load(self, value):
        if isinstance(value, Decimal):
            if value % 1 == 0:
                return int(value)
            else:
                return float(value)
        return value

register_type(NumberType)


class FloatType(TypeDefinition):

    """ Float values """
    data_type = float
    ddb_data_type = NUMBER

    def coerce(self, value, force):
        if not isinstance(value, float):
            # Auto-convert ints, longs, and Decimals
            if (isinstance(value, int) or isinstance(value, long) or
                    isinstance(value, Decimal)):
                return float(value)
            elif force:
                return float(value)
            else:
                raise TypeError()
        return value

    def ddb_load(self, value):
        return float(value)

register_type(FloatType)


class IntType(TypeDefinition):

    """ Integer values (includes longs) """
    data_type = int
    ddb_data_type = NUMBER

    def coerce(self, value, force):
        if not (isinstance(value, int) or isinstance(value, long)):
            if force:
                new_val = int(value)
                if isinstance(value, float) or isinstance(value, Decimal):
                    if new_val != value:
                        raise ValueError("Refusing to convert "
                                         "%s to int! Results in data loss!"
                                         % repr(value))
                return new_val
            else:
                raise TypeError()
        return value

    def ddb_load(self, value):
        return int(value)

register_type(IntType)


class DecimalType(TypeDefinition):

    """
    Numerical values that use Decimal in the application layer.

    This should be used if you want to work with floats but need the additional
    precision of the Decimal type.

    """

    data_type = Decimal
    ddb_data_type = NUMBER

    def coerce(self, value, force):
        if not isinstance(value, Decimal):
            if force:
                # Python 2.6 can't convert directly from float to Decimal
                if (isinstance(value, float) and
                        not hasattr(Decimal, 'from_float')):
                    return float_to_decimal(value)
                return Decimal(value)
            else:
                raise TypeError()
        return value

register_type(DecimalType)


class BoolType(TypeDefinition):

    """ Booleans, backed by a Dynamo Number """

    data_type = bool
    ddb_data_type = NUMBER

    def __init__(self):
        super(BoolType, self).__init__()
        self.allowed_filters = set(['eq', 'ne'])

    def coerce(self, value, force):
        if not isinstance(value, bool):
            if force:
                return bool(value)
            else:
                raise TypeError()
        return value

    def ddb_dump(self, value):
        return int(value)

    def ddb_load(self, value):
        return bool(value)

register_type(BoolType)


class StringType(TypeDefinition):

    """ String values, stored as unicode """
    data_type = unicode
    aliases = [STRING]
    ddb_data_type = STRING

    def coerce(self, value, force):
        if not isinstance(value, unicode):
            # Silently convert str to unicode using utf-8
            if isinstance(value, str):
                return value.decode('utf-8')
            if force:
                return unicode(value)
            else:
                raise TypeError()
        return value

register_type(StringType)


class BinaryType(TypeDefinition):

    """ Binary strings, stored as a str """
    data_type = str
    aliases = [BINARY, Binary, bytes]
    ddb_data_type = BINARY

    def coerce(self, value, force):
        if not isinstance(value, str):
            # Silently convert unicode to str using utf-8
            if isinstance(value, unicode):
                return value.encode('utf-8')
            if force:
                return str(value)
            else:
                raise TypeError()
        return value

    def ddb_dump(self, value):
        return Binary(value)

    def ddb_load(self, value):
        return value.value

register_type(BinaryType)


class DictType(TypeDefinition):

    """ Dict types, stored as a json string """
    data_type = dict
    ddb_data_type = STRING
    mutable = True

    def __init__(self):
        super(DictType, self).__init__()
        self.allowed_filters = set()

    def coerce(self, value, force):
        if not isinstance(value, dict):
            if force:
                return dict(value)
            else:
                raise TypeError()
        return value

    def ddb_dump(self, value):
        return json.dumps(value)

    def ddb_load(self, value):
        return json.loads(value)

register_type(DictType)


class ListType(TypeDefinition):

    """ List types, stored as a json string """
    data_type = list
    ddb_data_type = STRING
    mutable = True

    def __init__(self):
        super(ListType, self).__init__()
        self.allowed_filters = set()

    def coerce(self, value, force):
        if not isinstance(value, list):
            if force:
                return list(value)
            else:
                raise TypeError()
        return value

    def ddb_dump(self, value):
        return json.dumps(value)

    def ddb_load(self, value):
        return json.loads(value)

register_type(ListType)


class DateTimeType(TypeDefinition):

    """ Datetimes, stored as a unix timestamp """
    data_type = datetime.datetime
    ddb_data_type = NUMBER

    def ddb_dump(self, value):
        return float(value.strftime('%s.%f'))

    def ddb_load(self, value):
        return datetime.datetime.fromtimestamp(value)

register_type(DateTimeType)


class DateType(TypeDefinition):

    """ Dates, stored as timestamps """
    data_type = datetime.date
    ddb_data_type = NUMBER

    def ddb_dump(self, value):
        return int(value.strftime('%s'))

    def ddb_load(self, value):
        return datetime.date.fromtimestamp(value)

register_type(DateType)


class Key(boto.s3.key.Key):

    """ Subclass of boto S3 key that adds equality operators """

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return self.key == getattr(other, 'key', None)

    def __ne__(self, other):
        return not self.__eq__(other)

    def copy_data_from_key(self, key):
        """ Copy the data from a boto Key """
        for attr in ('name', 'metadata', 'cache_control', 'content_type',
                     'content_encoding', 'content_disposition',
                     'content_language', 'filename', 'etag', 'is_latest',
                     'last_modified', 'owner', 'storage_class', 'path', 'resp',
                     'mode', 'size', 'version_id', 'source_version_id',
                     'delete_marker', 'encrypted', 'ongoing_restore',
                     'expiry_date', 'local_hashes'):
            setattr(self, attr, getattr(key, attr))


class S3Type(TypeDefinition):

    """
    Store a link to an S3 key

    Parameters
    ----------
    bucket : str
        The name of the S3 bucket
    scheme : str, optional
        The name of the scheme to use to connect to S3 if the bucket is a
        string. (default 'default'). See :meth:`~.set_scheme` for more
        information.

    """
    ddb_data_type = STRING
    mutable = True
    SCHEMES = {
        'default': {},
    }
    _connections = {}

    def __init__(self, bucket, scheme='default'):
        super(S3Type, self).__init__()
        self._bucket_name = bucket
        self._bucket = None
        self._scheme = scheme

    @property
    def bucket(self):
        """ Getter for S3 bucket """
        if self._bucket is None:
            self._bucket = self._get_bucket(self._bucket_name, self._scheme)
        return self._bucket

    @classmethod
    def _get_bucket(cls, bucket, scheme):
        """ Get a connection to an S3 bucket """
        if scheme not in cls.SCHEMES:
            raise KeyError("Could not find S3 connection scheme '%s'. "
                           "Have you registered it with S3Type.set_scheme?"
                           % scheme)
        if scheme not in cls._connections:
            cls._connections[scheme] = boto.connect_s3(**cls.SCHEMES[scheme])
        return cls._connections[scheme].get_bucket(bucket, validate=False)

    @classmethod
    def set_scheme(cls, name, **kwargs):
        """
        Register a S3 connection scheme

        The connection scheme is a collection of keyword arguments that will be
        passed to :meth:`~boto.connect_s3` when creating a S3 connection.

        Parameters
        ----------
        name : str
            The name of the scheme. If the name is 'default', then that scheme
            will be used when no scheme is explicitly passed to the
            constructor.
        **kwargs : dict
            All keyword arguments

        """
        cls.SCHEMES[name] = kwargs

    def coerce(self, value, force):
        """ S3Type will auto-coerce string types """
        if not isinstance(value, Key):
            # silently convert boto Keys to our subclass
            if isinstance(value, boto.s3.key.Key):
                key = Key(self.bucket)
                key.copy_data_from_key(value)
                return key
            # Silently convert strings
            if isinstance(value, basestring):
                key = Key(self.bucket)
                key.key = value
                return key
            else:
                raise TypeError()
        return value

    def ddb_dump(self, value):
        return value.key

    def ddb_load(self, value):
        key = Key(self.bucket)
        key.key = value
        return key



""" Setup file """
import os
import sys

from setuptools import setup, find_packages


HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, 'README.rst')).read()
CHANGES = open(os.path.join(HERE, 'CHANGES.rst')).read()

REQUIREMENTS = [
    'boto>=2.23.0',
]

TEST_REQUIREMENTS = [
    'nose',
    'mock',
    'httpretty',
    'moto',
]

if sys.version_info[:2] < (2, 7):
    TEST_REQUIREMENTS.extend(['unittest2'])

if __name__ == "__main__":
    setup(
        name='flywheel',
        version='0.1.1',
        description="Object mapper for Amazon's DynamoDB",
        long_description=README + '\n\n' + CHANGES,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Topic :: Database',
        ],
        author='Steven Arcangeli',
        author_email='steven@highlig.ht',
        url='http://flywheel.readthedocs.org/',
        keywords='aws dynamo dynamodb orm odm',
        include_package_data=True,
        packages=find_packages(exclude=('tests',)),
        entry_points={
            'nose.plugins': [
                'dynamolocal=flywheel.tests:DynamoLocalPlugin',
            ],
        },
        install_requires=REQUIREMENTS,
        tests_require=REQUIREMENTS + TEST_REQUIREMENTS,
    )



""" Query and Scan builders """
from boto.dynamodb2.table import Table

from .fields import Condition


class Query(object):

    """
    An object used to query dynamo tables

    See the :class:`.Engine` for query examples

    Parameters
    ----------
    engine : :class:`.Engine`
    model : class
        Subclass of :class:`~flywheel.models.Model`

    """

    def __init__(self, engine, model):
        self.engine = engine
        self.model = model
        self.condition = Condition()

    @property
    def table(self):
        """ Shortcut to access dynamo table """
        return Table(self.model.meta_.ddb_tablename,
                     connection=self.engine.dynamo)

    def gen(self, desc=False, consistent=False, attributes=None):
        """
        Return the query results as a generator

        Parameters
        ----------
        desc : bool, optional
            Return results in descending order (default False)
        consistent : bool, optional
            Force a consistent read of the data (default False)
        attributes : list, optional
            List of fields to retrieve from dynamo. If supplied, gen() will
            iterate over boto ResultItems instead of model objects.

        """
        kwargs = self.condition.query_kwargs(self.model)
        if attributes is not None:
            kwargs['attributes'] = attributes
        kwargs['reverse'] = not desc
        kwargs['consistent'] = consistent
        results = self.table.query(**kwargs)
        for result in results:
            if attributes is not None:
                yield result
            else:
                yield self.model.ddb_load_(self.engine, result)

    def __iter__(self):
        for item in self.gen():
            yield item

    def all(self, desc=False, consistent=False, attributes=None):
        """
        Return the query results as a list

        Parameters
        ----------
        desc : bool, optional
            Return results in descending order (default False)
        consistent : bool, optional
            Force a consistent read of the data (default False)
        attributes : list, optional
            List of fields to retrieve from dynamo. If supplied, returns boto
            ResultItems instead of model objects.

        """
        return list(self.gen(desc=desc, consistent=consistent,
                             attributes=attributes))

    def first(self, desc=False, consistent=False, attributes=None):
        """
        Return the first result of the query, or None if no results

        Parameters
        ----------
        desc : bool, optional
            Return results in descending order (default False)
        consistent : bool, optional
            Force a consistent read of the data (default False)
        attributes : list, optional
            List of fields to retrieve from dynamo. If supplied, returns boto
            ResultItems instead of model objects.

        """
        self.limit(1)
        for result in self.gen(desc=desc, consistent=consistent,
                               attributes=attributes):
            return result
        return None

    def one(self, consistent=False, attributes=None):
        """
        Return the result of the query. If there is not exactly one result,
        raise a ValueError

        Parameters
        ----------
        consistent : bool, optional
            Force a consistent read of the data (default False)
        attributes : list, optional
            List of fields to retrieve from dynamo. If supplied, returns boto
            ResultItems instead of model objects.

        Raises
        ------
        exc : ValueError
            If there is not exactly one result

        """
        self.limit(2)
        items = self.all(consistent=consistent, attributes=attributes)
        if len(items) > 1:
            raise ValueError("More than one result!")
        elif len(items) == 0:
            raise ValueError("Expected one result!")
        return items[0]

    def limit(self, count):
        """ Limit the number of query results """
        self.condition &= Condition.construct_limit(count)
        return self

    def index(self, name):
        """ Use a specific local or global index for the query """
        self.condition &= Condition.construct_index(name)
        return self

    def delete(self):
        """ Delete all items that match the query """
        meta = self.model.meta_
        attrs = [meta.hash_key.name]
        if meta.range_key is not None:
            attrs.append(meta.range_key.name)
        results = self.gen(attributes=attrs)
        return self.engine.delete_key(self.model, results)

    def filter(self, *conditions, **kwargs):
        """
        Add a Condition to constrain the query

        Notes
        -----
        The conditions may be passed in as positional arguments:

        .. code-block:: python

            engine.query(User).filter(User.id == 12345)

        Or they may be passed in as keyword arguments:

        .. code-block:: python

            engine.query(User).filter(firstname='Monty', lastname='Python')

        The limitations of the keyword method is that you may only create
        equality conditions. You may use both types in a single filter:

        .. code-block:: python

            engine.query(User).filter(User.num_friends > 10, name='Monty')

        """
        for condition in conditions:
            self.condition &= condition
        for key, val in kwargs.iteritems():
            field = self.model.meta_.fields.get(key)
            if field is not None:
                self.condition &= (field == val)
            else:
                self.condition &= (self.model.field_(key) == val)

        return self


class Scan(Query):

    """
    An object used to scan dynamo tables

    scans are like Queries except they don't use indexes. This means they
    iterate over all data in the table and are SLOW

    Parameters
    ----------
    engine : :class:`.Engine`
    model : class
        Subclass of :class:`~flywheel.models.Model`

    """

    def gen(self, attributes=None, desc=False, consistent=False):
        if desc:
            raise ValueError("Cannot order scan() results")
        if consistent:
            raise ValueError("Cannot force consistent read on scan()")
        kwargs = self.condition.scan_kwargs()

        if attributes is not None:
            kwargs['attributes'] = attributes
        results = self.table.scan(**kwargs)
        for result in results:
            if attributes is not None:
                yield result
            else:
                yield self.model.ddb_load_(self.engine, result)

    def index(self, name):
        raise TypeError("Scan cannot use an index!")
